# ATtila
# Developed by Christian Visintin
# 
# MIT License
# Copyright (c) 2019 Christian Visintin
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

from .atsession import ATSession
from .atcommand import ATCommand
from .esk import ESKValue, ESK
from .atscriptparser import ATScriptParser
from .exceptions import ATScriptNotFound, ATScriptSyntaxError, ATSerialPortError, ATREUninitializedError, ATRuntimeError
from .atcommunicator import ATCommunicator
from .virtual.atvirtualcommunicator import ATVirtualCommunicator

from os import environ, system
from time import sleep

class ATRuntimeEnvironment(object):
  """
  This class represents the AT Runtime Environment which is the main component of
  ATtila, and the only one that should be instanced by the final user.
  The ATRE takes care of handling the communication between the session and the device and
  of the command execution flow
  """

  def __init__(self, abort_on_failure = True):
    """
    Class constructor. Instantiates a new :class:`.ATRuntimeEnvironment.` object with the provided parameters.
    
    :param abort_on_failure
    :type abort_on_failure bool
    """
    self.__session = ATSession()
    self.__communicator = ATCommunicator(None, None)
    self.__script_parser = ATScriptParser()
    self.__esks = []
    self.__virtual_communicator = False
    #ES Params
    self.__aof = abort_on_failure
    self.__current_command = 0

  @property
  def aof(self):
    return self.__aof

  def configure_communicator(self, serial_port, baud_rate, timeout = None, line_break ="\r\n"):
    """
    Configure ATRE communicator
    
    :param serial_port: Serial port for the communicator
    :param baud_rate: Baud rate used
    :param timeout: default timeout for commands
    :param line_break: line break used by the device
    :type serial_port: String
    :type baud_rate: int
    :type timeout: int
    :type line_break: String
    """
    if self.__communicator: #If device is open, close device
      if self.__communicator.is_open():
        self.__communicator.close()
    if self.__virtual_communicator:
      self.__communicator = ATCommunicator(serial_port, baud_rate)
    self.__virtual_communicator = False
    self.__communicator.serial_port = serial_port
    self.__communicator.baud_rate = baud_rate
    self.__communicator.default_timeout = timeout
    self.__communicator.line_break = line_break

  def configure_virtual_communicator(self, serial_port, baud_rate, timeout = None, line_break = "\r\n", read_callback = None, write_callback = None, in_waiting_callback = None):
    """
    Configure ATRE Virtual communicator

    :param serial_port: Serial port for the communicator
    :param baud_rate: Baud rate used
    :param timeout: default timeout for commands
    :param line_break: line break used by the device
    :param read_callback (optional): Specify a read function to call to read using the virtual communicator
    :param write_callback (optional): Specifiy a write funtion to call to write using the virtual communicator
    :param in_waiting_callback (optional): Specify a in waiting function to call
    :type serial_port: String
    :type baud_rate: int
    :type timeout: int
    :type line_break: String
    :type read_callback: function which returns string and takes nbytes as argument, if nbytes is -1 returns all lines
    :type write_callback: function which takes string and raises VirtualSerialException
    :type in_waiting_callback: function which returns True if there are data available to read
    """
    if self.__communicator: #If device is open, close device
      if self.__communicator.is_open():
        self.__communicator.close()
    self.__virtual_communicator = True
    self.__communicator = ATVirtualCommunicator(serial_port, baud_rate, timeout, line_break, read_callback, write_callback, in_waiting_callback)

  def init_session(self, commands):
    """
    Initialize a new ATSession

    :param commands
    :type commands: Array of ATCommands
    """
    self.__session.reset()
    for command in commands:
      self.__session.add_command(command)

  def set_ESKs(self, esks):
    """
    Set ATRE Environment Setup Keywords
    The previous ESKs will be overwritten, the newer will be associated to the current session

    :param esks
    :type esks: tuple of (ESKValue, execution_index)
    """
    self.__esks = esks

  def parse_ATScript(self, script_file):
    """
    Parse an AT Script file

    :param script: Script content
    :type script_file: String
    :raises ATScriptSyntaxError, ATScriptNotFound
    """
    try:
      parse_result = self.__script_parser.parse_file(script_file)
    except ATScriptNotFound as err:
      raise err
    except ATScriptSyntaxError as err:
      raise err
    commands = parse_result[0]
    esks = parse_result[1]
    self.init_session(commands)
    self.set_ESKs(esks)

  def add_command(self, command):
    """
    Add an ATCommand to the ATSession

    :param command: at command to add
    :type command: ATCommand
    :returns bool
    """
    return self.__session.add_command(command)

  def run(self):
    """
    Starts and run current ATSession

    :returns List of ATResponse
    :raises ATSerialPortError, ATRuntimeError, ATREUninitializedError
    """
    #Open serial port to initialize communication with device
    if not self.__communicator or not self.__session:
      raise ATREUninitializedError("AT Runtime Environment is not initialized")
    try:
      self.open_serial()
    except ATSerialPortError as err:
      raise err
    response_list = []
    while self.__session.get_next_command(): #For each command execute it
      response = self.exec_next()
      response_list.append(response)
      #Proceed with the next command (or its doppelganger maybe...)
    #Close serial
    try:
      self.close_serial()
    except ATSerialPortError as err:
      raise err
    return response_list

  def exec(self, command):
    """
    Execute in the current session a command or a ESK.
    The command has to be parsed by th eATScriptParser before being executed.
    An ATResponse is returned if was a command, otherwise None
    This method doesn't open or close the serial
    
    :param command
    :type command String
    :returns ATResponse or None
    :raises ATScriptSyntaxError, ATSerialPortError, ATREUninitializedError, ATRuntimeError
    """
    #Try to parse command
    try:
      parse_result = self.__script_parser.parse(command)
    except ATScriptSyntaxError as err:
      raise err
    commands = parse_result[0]
    esks = parse_result[1]
    if len(commands) > 0:
      command = commands[0]
      if not self.__session:
        raise ATREUninitializedError("Session is not initialized")
      if not self.__communicator.serial_port:
        raise ATREUninitializedError("Communicator is not initialized")
      #Clear commands in order to prevent conflicts
      self.__session.clear_commands()
      #Add command to session
      self.__session.add_command(command)
      atcmd = self.__session.get_next_command()
      #Delay
      if atcmd.delay:
        sleep(atcmd.delay / 1000)
      #Execute command on device
      response, execution_time = self.__communicator.exec(atcmd.command, atcmd.timeout)
      #Validate response
      response = self.__session.validate_response(response, execution_time)
      if self.__session.last_command_failed and not atcmd.doppel_ganger and self.__aof:
        raise ATRuntimeError("Command '%s' got a bad response: '%s' (and hasn't any doppelganger)!" % (atcmd.command, response.full_response))
      return response
    elif len(esks) > 0:
      #Process ESK
      esk = esks[0]
      if not self.__process_ESK(esk[0]):
        raise ATRuntimeError("ESK %s failed" % esk[0].keyword)
      return None
    else:
      return None

  def exec_next(self):
    """
    Execute next command (It doesn't open/close the serial port)

    :returns ATResponse (None if there's no command to execute)
    :raises ATSerialPortError, ATRuntimeError
    """
    #Before executing command, check if an ESK has to be executed
    esks = [i[0] for i in self.__esks if i[1] == self.__current_command]
    for esk in esks:
      if not self.__process_ESK(esk) and self.__aof:
        raise ATRuntimeError("Runtime Error while processing ESK (%s %s)" % (esk.keyword, esk.value))
    #Then remove already executed esks esks
    self.__esks = [i for i in self.__esks if i[1] != self.__current_command]
    #Get next command
    next_command = self.__session.get_next_command()
    if not next_command:
      return None
    #Delay
    if next_command.delay:
      sleep(next_command.delay / 1000)
    #Send command to communicator
    try:
      response, execution_time = self.__communicator.exec(next_command.command, next_command.timeout)
    except ATSerialPortError as err:
      raise err
    #Validate response
    response = self.__session.validate_response(response, execution_time)
    #Check if last command failed; if it hasn't a doppelganger and abort on failure is True, then raise RuntimeError
    if self.__session.last_command_failed and not next_command.doppel_ganger and self.__aof:
      raise ATRuntimeError("Command '%s' got a bad response: '%s' (and hasn't any doppelganger)!" % (next_command.command, response.full_response))
    if not self.__session.last_command_failed:
      self.__current_command += 1
    return response

  def open_serial(self):
    """
    Open Serial port

    :raises ATSerialError, ATREUninitializedError
    """
    if not self.__communicator.serial_port:
      raise ATREUninitializedError("Communicator is not initialized")
    if self.__communicator.is_open():
      return
    try:
      self.__communicator.open()
    except ATSerialPortError as err:
      raise err

  def close_serial(self):
    """
    Close Serial Port

    :raises ATSerialError, ATREUninitializedError
    """
    if not self.__communicator.serial_port:
      raise ATREUninitializedError("Communicator is not initialized")
    if not self.__communicator.is_open():
      return
    try:
      self.__communicator.close()
    except ATSerialPortError as err:
      raise err

  def get_session_value(self, key):
    """
    Try to get a value from the current session storage
    If key doesn't exist, KeyError is raised

    :param key
    :type key: String
    :returns any
    :raises KeyError
    """
    try:
      self.__session.get_session_value(key)
    except KeyError as err:
      raise err

  def __process_ESK(self, esk):
    """
    Process an environment setup keyword

    :param esk
    :type esk: ESKValue
    :returns bool
    """
    if not esk:
      return False
    elif esk.keyword is ESK.DEVICE:
      #Check if serial is opened
      if self.__communicator:
        if self.__communicator.is_open():
          self.__communicator.close()
      self.configure_communicator(esk.value, self.__communicator.baud_rate)
      if self.__communicator.serial_port and self.__communicator.baud_rate:
        try:
          self.__communicator.open()
          return True
        except ATSerialPortError:
          return False
    elif esk.keyword is ESK.BAUDRATE:
      #Check if serial is opened
      if self.__communicator:
        if self.__communicator.is_open():
          self.__communicator.close()
      self.configure_communicator(self.__communicator.serial_port, esk.value)
      if self.__communicator.serial_port and self.__communicator.baud_rate:
        try:
          self.__communicator.open()
          return True
        except ATSerialPortError:
          return False
    elif esk.keyword is ESK.TIMEOUT:
      self.__communicator.default_timeout = esk.value
    elif esk.keyword is ESK.BREAK:
      self.__communicator.line_break = esk.value
    elif esk.keyword is ESK.AOF:
      self.__aof = esk.value
    elif esk.keyword is ESK.SET:
      #Set a session value
      if self.__session:
        self.__session.set_session_value(esk.value[0], esk.value[1])
    elif esk.keyword is ESK.GETENV:
      #Try to get key from environment
      try:
        env_value = environ[esk.value]
        self.__session.set_session_value(esk.value, env_value)
      except KeyError:
        return False
    elif esk.keyword is ESK.PRINT:
      #Replace session values
      to_out = self.__session.replace_session_keys(esk.value)
      print(to_out)
    elif esk.keyword in ESK.EXEC:
      rc = system(esk.value)
      if rc != 0:
        return False
    else:
      return False
    return True



# ATtila
# Developed by Christian Visintin
# 
# MIT License
# Copyright (c) 2019 Christian Visintin
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

from .exceptions import ATScriptNotFound, ATScriptSyntaxError
from .atcommand import ATCommand
from .esk import ESK, ESKValue

class ATScriptParser(object):
  """
  This class represents an AT script parser, which is the component which purpose is to parse an ATScript.
  """

  def parse(self, script):
    """
    Parse an ATScript

    :param script: lines of at script
    :type script: String
    :returns Tuple of array of ATCommand and array of tuple of (ESKValue, execution index)
    :raises ATScriptSyntaxError
    """
    result = ([], [])
    commands = result[0]
    esks = result[1]
    execution_index = 0
    line_no = 0
    #Split script into rows
    for row in script.splitlines():
      #Increment line number
      line_no += 1
      if not row: #Empty row
        continue
      if row.startswith("#"): #Is comment
        continue
      eks, error = self.__parse_esk(row)
      if eks:
        esks.append((eks, execution_index))
      elif error: #If error is set, it means line is EKS, but has invalid syntax
        raise ATScriptSyntaxError("Syntax error at line %d: %s (%s)" % (line_no, error, row))
      else: #Otherwise, if both are None, it is a command
        #Try as ommand
        command, error = self.__parse_command(row)
        if command:
          commands.append(command)
          #Increment execution_index
          execution_index += 1
        elif error:
          raise ATScriptSyntaxError("Syntax error at line %d: %s (%s)" % (line_no, error, row))
        else:
          raise ATScriptSyntaxError("Syntax error at line %d: %s -- Don't know how to interpret this line, sorry..." % (line_no, row))
    return result

  def parse_file(self, file_path):
    """
    Parse an ATScript file

    :param file_path: path of the ATScript file
    :type file_path: String
    :returns Tuple of array of ATCommand and array of tuple of (ESKValue, execution index)
    :raises ATScriptNotFound, ATScriptSyntaxError
    """
    try:
      hnd = open(file_path)
      script = hnd.read()
      hnd.close()
      return self.parse(script)
    except IOError as err:
      raise ATScriptNotFound(err)
    except ATScriptSyntaxError as err:
      raise err

  def __parse_esk(self, row):
    """
    Parse a row, which is possibly an ESK

    :param row: ESK row
    :type row: String
    :returns (ESKValue, String): A tuple of ESKValue and an error string. Error is different from None only if ESK is valid. Both are None if it's not an ESK.
    """
    esk_value = None
    error = None
    line_tokens = row.split()
    if len(line_tokens) == 0:
      error = "Empty row"
      return (esk_value, error)
    esk_str = line_tokens[0]
    if len(line_tokens) > 1:
      esk_attr = " ".join(line_tokens[1:])
    else:
      esk_attr = None
    esk = ESK.get_esk_from_string(esk_str)
    if not esk:
      #Not an ESK
      return (esk_value, error) #None, None
    #Eval attributes
    esk_value = ESK.to_ESKValue(esk, esk_attr)
    if not esk_value:
      error = "Invalid attributes"
    return (esk_value, error)

  def __parse_command(self, row):
    """
    Parse a row, which is possibly an AT command

    :param row: command row
    :type row: String
    :returns (ATCommand, String): A tuple of ATCommand and an error string. Error is different from None only if command syntax is wrong. Both are None if it's not an AT command.
    """
    command = None
    error = None
    command_tokens = row.split(";;")
    if len(command_tokens) == 0:
      error = "Empty row"
      return (command, error)
    atcommand = command_tokens[0]
    expected_response = None
    delay = 0
    timeout = None
    collectables = None
    doppelganger = None
    doppelganger_response = None
    has_doppelganger = False
    if len(command_tokens) > 1: #Expected response
      if command_tokens[1]:
        expected_response = command_tokens[1]
    if len(command_tokens) > 2: #Delay
      if command_tokens[2]:
        try:
          delay = int(command_tokens[2])
        except ValueError:
          error = "Delay is not a number"
          return (command, error)
    if len(command_tokens) > 3: #Timeout
      if command_tokens[3]:
        try:
          timeout = int(command_tokens[3])
        except ValueError:
          error = "Timeout is not a number"
          return (command, error)
    if len(command_tokens) > 4: #Collectables
      if command_tokens[4]:
        try:
          collectables = eval(command_tokens[4])
        except NameError as err:
          error = "Collectables has invalid syntax (%s)" % err
          return (command, error)
        except SyntaxError as err:
          error = "Collectables has invalid syntax (%s)" % err
          return (command, error)
    if len(command_tokens) > 5: #Doppelganger
      if command_tokens[5]:
        has_doppelganger = True
    if len(command_tokens) > 6: #Doppelganger response
      if command_tokens[6]:
        doppelganger_response = command_tokens[6]
    if has_doppelganger:
      #Instance doppelganger
      doppelganger = ATCommand(command_tokens[5], doppelganger_response, timeout, delay, collectables, None)
    #Instance new AT command
    command = ATCommand(atcommand, expected_response, timeout, delay, collectables, doppelganger)
    return (command, error)



# ATtila
# Developed by Christian Visintin
# 
# MIT License
# Copyright (c) 2019 Christian Visintin
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

from .atresponse import ATResponse

class ATCommand(object):
  """
  This class represents an AT command
  """
  
  def __init__(self, cmd, exp_response = None, tout = None, delay = 0, collectables = None, dganger = None):
    """
    Class constructor. Instantiates a new :class:`.ATCommand.` object with the provided parameters.
    
    :param cmd: command to execute
    :param exp_response: expected response from command execution. a literal or a generic response can be provided
    :param tout (optional): command timeout execution in seconds.
    :param delay (optional): delay in milliseconds before command execution
    :param collectables (optional): values to store from response. Follow collectables syntax as specified in ATtila documentation
    :param dganger (optional): doppelganger command associated to this command (command to execute in case of this command fails)
    :type cmd: string
    :type ATResponse: string
    :type exp_respose: string
    :type tout: int
    :type delay: int
    :type collectables: list of string
    :type dganger: ATCommand
    """
    self._command = cmd
    self._expected_response = exp_response
    self._timeout = tout
    self._delay = delay
    self.collectables = collectables
    self.doppel_ganger = dganger
    self._response = None

  @property
  def command(self):
    return self._command

  @command.setter
  def command(self, cmd):
    self._command = cmd

  @property
  def expected_response(self):
    return self._expected_response

  @expected_response.setter
  def expected_response(self, exp_response):
    self._expected_response = exp_response

  @property
  def response(self):
    return self._response

  @response.setter
  def response(self, response):
    if isinstance(response, ATResponse):
      self._response = response
    else:
      self._response = None

  @property
  def timeout(self):
    return self._timeout

  @timeout.setter
  def timeout(self, tout):
    if tout > 1:
      self._timeout = tout
    else:
      self._timeout = 1

  @property
  def delay(self):
    return self._delay

  @delay.setter
  def delay(self, delay):
    if delay > 0:
      self._delay = delay
    else:
      self._delay = 0

  @property
  def collectables(self):
    return self._collectables

  @collectables.setter
  def collectables(self, collectables):
    if type(collectables) == list:
      self._collectables = collectables
    else:
      self._collectables = None

  @property
  def doppel_ganger(self):
    return self._doppel_ganger

  @doppel_ganger.setter
  def doppel_ganger(self, dganger):
    if isinstance(dganger, ATCommand):
      self._doppel_ganger = dganger
    else:
      self._doppel_ganger = None



# ATtila
# Developed by Christian Visintin
# 
# MIT License
# Copyright (c) 2019 Christian Visintin
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

from enum import Enum

class ESKValue(object):
  """
  This class represents an Environment Setup Keyword value
  """

  def __init__(self, keyword, value):
    """
    Class constructor. Instantiates a new :class:`.ESKValue.` object with the provided paramters

    :param keyword: keyword type
    :param value associated
    :type keyword: ESK
    :type value: Any
    """
    self._keyword = keyword
    self._value = value

  @property
  def keyword(self):
    return self._keyword

  @keyword.setter
  def keyword(self, kw):
    self._keyword = kw

  @property
  def value(self):
    return self._value

  @value.setter
  def value(self, val):
    self._value = val


class ESK(Enum):
  """
  This class represents an Environment Setup Keyword
  """
  DEVICE = 0
  BAUDRATE = 1
  TIMEOUT = 2
  BREAK = 3
  AOF = 4
  SET = 5
  GETENV = 6
  PRINT = 7
  EXEC = 8

  @staticmethod
  def get_esk_from_string(esk_string):
    """
    Get ESK enum value from its string representation

    :param esk_string
    :type esk_string: String
    :returns ESK or None if invalid
    """
    if esk_string == "DEVICE":
      return ESK.DEVICE
    elif esk_string == "BAUDRATE":
      return ESK.BAUDRATE
    elif esk_string == "TIMEOUT":
      return ESK.TIMEOUT
    elif esk_string == "BREAK":
      return ESK.BREAK
    elif esk_string == "AOF":
      return ESK.AOF
    elif esk_string == "SET":
      return ESK.SET
    elif esk_string == "GETENV":
      return ESK.GETENV
    elif esk_string == "PRINT":
      return ESK.PRINT
    elif esk_string == "EXEC":
      return ESK.EXEC
    else:
      return None

  @staticmethod
  def to_ESKValue(esk, attr):
    """
    Check if attributes for this esk have a valid syntax

    :param esk
    :param attr
    :type esk: ESK
    :type attr: String
    :returns ESKValue (None in case of errors)
    """
    if not esk:
      return False
    elif esk is ESK.DEVICE:
      if attr:
        return ESKValue(esk, attr)
      else:
        return None
    elif esk is ESK.BAUDRATE:
      try:
        baud = int(attr)
        return ESKValue(esk, baud)
      except ValueError: #NaN
        return None
    elif esk is ESK.TIMEOUT:
      try:
        timeout = int(attr)
        return ESKValue(esk, timeout)
      except ValueError: #NaN
        return None
    elif esk is ESK.BREAK:
      if attr == "LF":
        return ESKValue(esk, "\n")
      elif attr == "CRLF":
        return ESKValue(esk, "\r\n")
      elif attr == "CR":
        return ESKValue(esk, "\r")
      elif attr == "NONE":
        return ESKValue(esk, None)
      else:
        return None
    elif esk is ESK.AOF:
      check = attr.lower()
      if check == "true":
        return ESKValue(esk, True)
      elif check == "false":
        return ESKValue(esk, False)
      else:
        return None
    elif esk is ESK.SET:
      key_val = attr.split("=")
      if len(key_val) == 2:
        return ESKValue(esk, (key_val[0], key_val[1])) #Tuple of key and value
      else:
        return None
    elif esk is ESK.GETENV:
      if attr:
        return ESKValue(esk, attr)
      else:
        return None
    elif esk is ESK.PRINT:
      if attr:
        return ESKValue(esk, attr)
      else:
        return None
    elif esk is ESK.EXEC:
      if attr:
        return ESKValue(esk, attr)
      else:
        return None
    else:
      return None



#!/usr/bin/python3

from setuptools import find_packages, setup, Extension
from pathlib import Path
from os import path

# The directory containing this file
ROOT = Path(__file__).parent

README = (ROOT / "README.md").read_text()

setup(
  name='attila',
  version='1.1.0',
  description='Python module to communicate easily with modems and RF modules using AT commands',
  long_description=README,
  long_description_content_type="text/markdown",
  author='Christian Visintin',
  author_email='christian.visintin1997@gmail.com',
  url='https://github.com/ChristianVisintin/ATtila',
  license="MIT",
  python_requires='>=3.4',
  include_package_data = True,
  install_requires=[
    'pyserial>=3'
  ],
  entry_points={
    'console_scripts': [
      'attila = attila.__main__:main'
    ]
  },
  packages=find_packages(exclude=("tests",)),
  keywords=['AT commands', 'IOT', 'wireless', 'radio frequency', 'modem', 'rf modules'],
  classifiers=[
    'Development Status :: 5 - Production/Stable',
    'Intended Audience :: Developers',
    'Intended Audience :: Telecommunications Industry',
    'Intended Audience :: End Users/Desktop',
    'Topic :: Software Development :: Libraries',
    'Topic :: Home Automation',
    'Topic :: Communications :: Telephony',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3',
    'Operating System :: OS Independent',
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.4",
  ]
)



# ATtila
# Developed by Christian Visintin
# 
# MIT License
# Copyright (c) 2019 Christian Visintin
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

PROGRAM_NAME = "attila"

#Import ATtila
from attila.atre import ATRuntimeEnvironment
from attila.exceptions import ATREUninitializedError, ATRuntimeError, ATScriptNotFound, ATScriptSyntaxError, ATSerialPortError
#System imports
from sys import argv, exit, stdout
#Logging
import logging
#Getopt
from getopt import getopt, GetoptError
#Signals
from signal import signal, SIGTERM, SIGINT

USAGE = "Usage: %s [OPTION]... [FILE]\n\
  \n\
  With no FILE, run in interactive mode\n\
  \n\
  \t-p <device path>\tUse this device to communicate\n\
  \t-b <baud rate>\t\tUse the specified baudrate to communicate\n\
  \t-T <default timeout>\tUse the specified timeout as default to communicate\n\
  \t-B <break>\t\tUse the specified line break [CRLF, LF, CR, NONE] (Default: CRLF)\n\
  \t-A <True/False>\t\tAbort on failure (Default: True)\n\
  \t-L <logfile>\t\tEnable log and log to the specified log file (stdout is supported)\n\
  \t-l <loglevel>\t\tSpecify the log level (0: CRITICAL, 1: ERROR, 2: WARN, 3: INFO, 4: DEBUG) (Default: INFO\n\
  \t-v\t\t\tBe more verbose\n\
  \t-q\t\t\tBe quiet (print only PRINT ESKs and ERRORS)\n\
  \t-h\t\t\tShow this page\n\
  " % PROGRAM_NAME

LOG_LEVEL_DEBUG = 4
LOG_LEVEL_INFO = 3
LOG_LEVEL_WARN = 2
LOG_LEVEL_ERROR = 1
LOG_LEVEL_CRITICAL = 0

#Globals
sigterm_called = False
interactive_mode = True

def sigterm_handler(_signo, _stack_frame):
  """
  Handle sigterm (or sigint) setting sigterm_called to True
  """
  global interactive_mode
  global sigterm_called
  logging.warn("SIGTERM called")
  sigterm_called = True
  if interactive_mode:
    print("Press ENTER to QUIT")

def opt_error(message):
  """
  Function to call in case of an error while parsing options and terminates with exit code 1

  :param message
  :type message: String
  """
  print(message)
  print(USAGE)
  exit(1)

def get_log_level_from_option(log_level_int):
  """
  Get log level from attila option

  :param log_level_num
  :type int
  :returns logging level
  """
  if log_level_int == LOG_LEVEL_DEBUG:
    return logging.DEBUG
  elif log_level_int == LOG_LEVEL_INFO:
    return logging.INFO
  elif log_level_int == LOG_LEVEL_WARN:
    return logging.WARNING
  elif log_level_int == LOG_LEVEL_ERROR:
    return logging.ERROR
  elif log_level_int == LOG_LEVEL_CRITICAL:
    return logging.CRITICAL
  else:
    return logging.INFO


def main():
  global sigterm_called
  global interactive_mode
  #Options
  script_file = None
  device = None
  baud_rate = None
  default_timeout = 0
  line_break = None
  logfile = None
  log_level = LOG_LEVEL_INFO
  verbose = False
  quiet = False
  abort_on_failure = True
  to_stdout = False

  try:
    optlist, args = getopt(argv[1:], "p::b::T::B::L::l::Avqh")
    if args:
      interactive_mode = False
      script_file = args[0]
    for opt, arg in optlist:
      if opt == "-p":
        device = arg
      elif opt == "-b":
        try:
          baud_rate = int(arg)
        except ValueError:
          opt_error("Specified baud rate is not a number!")
      elif opt == "-T":
        try:
          default_timeout = int(arg)
        except ValueError:
          opt_error("Specified default timeout is not a number!")
      elif opt == "-B":
        line_break = arg
        #Verify line break
        if line_break == "CRLF":
          line_break = "\r\n"
        elif line_break == "LF":
          line_break = "\n"
        elif line_break == "CR":
          line_break = "\r"
        elif line_break == "NONE":
          line_break = None
        else:
          opt_error("Invalid line break '%s'" % line_break)
      elif opt == "-L":
        logfile = arg
      elif opt == "-l":
        try:
          log_level = int(arg)
          if log_level < LOG_LEVEL_CRITICAL or log_level > LOG_LEVEL_DEBUG:
            opt_error("Log level is out of range")
          else:
            log_level = get_log_level_from_option(log_level)
        except ValueError:
          opt_error("Log level is not a number")
      elif opt == "-A":
        try:
          abort_on_failure = eval(arg)
          if abort_on_failure != True and abort_on_failure != False:
            opt_error("Abort on failure has a bad value: '%s', but should be True or False" % abort_on_failure)
        except NameError:
          opt_error("Abort on failure has a bad value")
      elif opt == "-v":
        verbose = True
      elif opt == "-q":
        quiet = True
      elif opt == "-h":
        print(USAGE)
        exit(0)
      else:
        opt_error("Unkown option '%s'" % opt)
  except GetoptError as err:
    opt_error(err)

  #Set signal handlers listener
  signal(SIGTERM, sigterm_handler)
  signal(SIGINT, sigterm_handler)

  #Prepare logger if requested
  if logfile:
    if logfile == "stdout":
      to_stdout = True
      stdout_handler = logging.StreamHandler(stdout)
      stdout_handler.setFormatter(logging.Formatter("%(asctime)s [%(levelname)s]: %(message)s", datefmt="%Y-%m-%dT%H:%M:%S"))
      logging.getLogger().addHandler(stdout_handler)
    else:
      logging.basicConfig(filename=logfile, level=log_level, format="%(asctime)s [%(levelname)s]: %(message)s", datefmt="%Y-%m-%dT%H:%M:%S")
  else:
    logging.getLogger().disabled = True
  #Instance ATRuntime environment
  atrunenv = ATRuntimeEnvironment(abort_on_failure)
  #Configure serial
  if device and baud_rate:
    atrunenv.configure_communicator(device, baud_rate, default_timeout, line_break)
    logging.info("Setup communicator (device: %s, baud_rate: %d)" % (device, baud_rate))
    if verbose and not to_stdout:
      print("Setup communicator (device: %s, baud_rate: %d)" % (device, baud_rate))
    #Open serial
    try:
      atrunenv.open_serial()
      logging.info("Serial port opened (%s)" % device)
      if verbose and not to_stdout:
        print("Serial port opened")
    except ATSerialPortError as err:
      logging.error("Could not open serial port: %s" % err)
      if not to_stdout and not quiet:
        print("Could not open serial port: %s" % err)
    except ATREUninitializedError as err:
      logging.error("Uninitialized runtime environment: %s " % err)
      if not to_stdout and not quiet:
        print("Uninitialized runtime environment: %s" % err)
      exit(1)
  #Parse file if set
  if script_file:
    logging.debug("Trying to parse file %s..." % script_file)
    try:
      atrunenv.parse_ATScript(script_file)
    except ATScriptNotFound as err:
      logging.error("Could not find script file %s: %s" % (script_file, err))
      if not to_stdout:
        print("Could not find script file %s: %s" % (script_file, err))
      exit(1)
    except ATScriptSyntaxError as err:
      logging.error("Script Syntax error: %s" % err)
      if not to_stdout:
        print("Script syntax error: %s" % err)
      exit(1)
    #Execute script
    response = 1
    while response and not sigterm_called:
      try:
        response = atrunenv.exec_next()
        if not response:
          continue
        #Handle response
        if response.response and response.command:
          logging.info("%s (%d ms) >> %s" % (response.command.command, response.execution_time, response.response))
          if not to_stdout and not quiet:
            print("%s (%d ms) >> %s" % (response.command.command, response.execution_time, response.response))
        else: #Command failed (this snippet gets executed only if aof is false)
          logging.error("%s (%d ms) >> %s" % (response.command.command, response.execution_time, "\n".join(response.full_response)))
          if not to_stdout and not quiet:
            print("%s (%d ms) >> %s" % (response.command.command, response.execution_time, "\n".join(response.full_response)))
      except ATSerialPortError as err:
        logging.error("Serial Port error: %s" % err)
        if not to_stdout:
          print("Serial Port error: %s" % err)
        if atrunenv.aof:
          break
      except ATRuntimeError as err:
        logging.error("Runtime error: %s" % err)
        if not to_stdout:
          print("Runtime error: %s" % err)
        if atrunenv.aof:
          break
      except ATREUninitializedError as err:
        logging.error("Uninitialized runtime environment: %s " % err)
  else: #Interactive mode
    while not sigterm_called:
      command_line = input(">> ")
      if not command_line:
        continue
      #Parse and execute command
      try:
        response = atrunenv.exec(command_line)
        if response: #Was a command (otherwise was probably ESK)
          if response.response: #Command was successful
            logging.info("%s (%d ms) >> %s" % (response.command.command, response.execution_time, response.response))
            if not to_stdout and not quiet:
              print("<< %s (%d ms) >> %s" % (response.command.command, response.execution_time, response.response))
          else: #Command error
            logging.error("%s (%d ms) >> %s" % (response.command.command, response.execution_time, "\n".join(response.full_response)))
            if not to_stdout and not quiet:
              print("<< %s (%d ms) >> %s" % (response.command.command, response.execution_time, "\n".join(response.full_response)))
        else:
          logging.info("%s >> OK" % command_line)
          if not to_stdout and not quiet:
            print("%s >> OK" % command_line)
      except ATScriptSyntaxError as err:
        logging.error("Syntax error: %s" % err)
        if not to_stdout:
          print("Syntax error: %s" % err)
        continue
      except ATSerialPortError as err:
        logging.error("Serial Port error: %s" % err)
        if not to_stdout:
          print("Serial Port error: %s" % err)
      except ATRuntimeError as err:
        logging.error("Runtime error: %s" % err)
        if not to_stdout:
          print("Runtime error: %s" % err)
  #Close serial
  try:
    atrunenv.close_serial()
  except ATSerialPortError as err:
    logging.error("Could not close serial port: %s" % err)
    if not to_stdout and not quiet:
      print("Could not close serial port: %s" % err)
    exit(1)
  except ATREUninitializedError as err:
    logging.error("Couldn't close serial port, since device was not initialized")
    if verbose and not to_stdout:
      print("Couldn't close serial port, since device was not initialized")
  #Execution terminated
  logging.info("attila terminated with exit code 0")
  if not to_stdout and verbose:
    print("attila terminated with exit code 0")

if __name__ == "__main__":
  main()



# ATtila
# Developed by Christian Visintin
# 
# MIT License
# Copyright (c) 2019 Christian Visintin
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

class ATResponse(object):
  """
  This class represents an AT command response and provide access to the expected
  response format, the entire response and the command execution time (milliseconds)
  """

  def __init__(self, resp, fullresponse, command, executiontime = 0):
    """
    Class constructor. Instantiates a new :class:`.ATResponse.` object with the provided parameters.

    :param resp: main response from command (format/content as expected by the user)
    :param fullresponse: entire response received from command execution
    :param command: command associated to response
    :param executiontime: execution time of the command in milliseconds
    :type resp: string
    :type fullresponse: list of string
    :type command: ATCommand
    :type executiontime: int
    """
    self._response = resp
    self._full_response = fullresponse
    self._execution_time = executiontime
    self._command = command
    self._collectables = {}

  @property
  def response(self):
    return self._response

  @response.setter
  def response(self, response):
    self._response = response

  @property
  def full_response(self):
    return self._full_response

  @full_response.setter
  def full_response(self, full_response):
    self._full_response = full_response

  @property
  def command(self):
    return self._command

  @property
  def execution_time(self):
    return self._execution_time

  @execution_time.setter
  def execution_time(self, executiontime):
    if executiontime > 0:
      self._execution_time = executiontime
    else:
      self._execution_time = 0

  def add_collectable(self, key, value):
    """
    Add a collectable to the response collectables

    :param key: key of the collectable
    :param value: value of the collectable
    :type key: string
    :type value: any
    :returns void
    """

    self._collectables[key] = value
    return

  def get_collectable(self, key):
    """
    Returns the value associated to the collectable
    
    :param key: key of the collectable
    :type key: string
    :returns any
    """
    return self._collectables.get(key)
